# Generated by roxygen2: do not edit by hand

S3method(as_draws,CmdStanGQ)
S3method(as_draws,CmdStanMCMC)
S3method(as_draws,CmdStanMLE)
S3method(as_draws,CmdStanVB)
export(as_cmdstan_fit)
export(as_draws)
export(as_mcmc.list)
export(check_cmdstan_toolchain)
export(cmdstan_default_install_path)
export(cmdstan_default_path)
export(cmdstan_make_local)
export(cmdstan_model)
export(cmdstan_path)
export(cmdstan_version)
export(cmdstanr_example)
export(draws_to_csv)
export(eng_cmdstan)
export(install_cmdstan)
export(num_threads)
export(print_example_program)
export(read_cmdstan_csv)
export(read_sample_csv)
export(rebuild_cmdstan)
export(register_knitr_engine)
export(set_cmdstan_path)
export(set_num_threads)
export(write_stan_file)
export(write_stan_json)
export(write_stan_tempfile)
import(R6)
importFrom(posterior,as_draws)
