---
name: Bug report
about: Create a report to help us improve
title: ''
labels: bug
assignees: ''

---

**Describe the bug**
A description of what the bug is.

**To Reproduce**
Steps to reproduce the behavior.

**Expected behavior**
A description of what you expected to happen.

**Operating system**
Your operating system (e.g. mac os x 10.15, windows 10, etc.)

**CmdStanR version number**
Your CmdStanR version number (e.g. from `packageVersion("cmdstanr")`).

**Additional context**
Add any other context about the problem here.
