% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr.R
\name{eng_cmdstan}
\alias{eng_cmdstan}
\title{CmdStan knitr engine for Stan}
\usage{
eng_cmdstan(options)
}
\arguments{
\item{options}{(named list) Chunk options, as provided by \code{knitr} during
chunk execution.}
}
\description{
This provides a knitr engine for Stan, suitable for usage when attempting
to render Stan chunks and compile the model code within to an executable with
CmdStan. Use \code{\link[=register_knitr_engine]{register_knitr_engine()}} to make this the default engine for
\code{stan} chunks. See the vignette
\href{https://mc-stan.org/cmdstanr/articles/r-markdown.html}{R Markdown CmdStan Engine}
for an example.
}
\examples{
\dontrun{
knitr::knit_engines$set(stan = cmdstanr::eng_cmdstan)
}
}
