% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{stan_threads}
\alias{stan_threads}
\alias{num_threads}
\alias{set_num_threads}
\title{Set or get the number of threads used to execute Stan models}
\usage{
num_threads()

set_num_threads(num_threads)
}
\arguments{
\item{num_threads}{(positive integer) The number of threads to set.}
}
\value{
The value of the environment variable \code{STAN_NUM_THREADS}.
}
\description{
DEPRECATED. Please use the \code{threads_per_chain} argument when fitting the model.
}
\keyword{internal}
