% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit.R
\name{fit-method-code}
\alias{fit-method-code}
\alias{code}
\title{Return Stan code}
\usage{
code()
}
\value{
A character vector with one element per line of code.
}
\description{
Return Stan code
}
\examples{

\dontrun{
fit <- cmdstanr_example()
fit$code() # character vector
cat(fit$code(), sep = "\n") # pretty print
}

}
\seealso{
\code{\link{CmdStanMCMC}}, \code{\link{CmdStanMLE}}, \code{\link{CmdStanVB}}, \code{\link{CmdStanGQ}}
}
