% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{cmdstan_default_path}
\alias{cmdstan_default_path}
\title{cmdstan_default_path}
\usage{
cmdstan_default_path(old = FALSE, dir = NULL)
}
\arguments{
\item{old}{See \code{\link[=cmdstan_default_install_path]{cmdstan_default_install_path()}}.}

\item{dir}{Path to a custom install folder with CmdStan installations.}
}
\value{
Path to the CmdStan installation with the most recent release
version, or \code{NULL} if no installation found.
}
\description{
Returns the path to the installation of CmdStan with the most recent release
version.
}
\details{
For Windows systems with WSL CmdStan installs, if there are side-by-side WSL
and native installs with the same version then the WSL is preferred.
Otherwise, the most recent release is chosen, regardless of whether it is
native or WSL.
}
\keyword{internal}
