% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/csv.R
\name{read_sample_csv}
\alias{read_sample_csv}
\title{Read CmdStan CSV files from sampling into \R}
\usage{
read_sample_csv(files, variables = NULL, sampler_diagnostics = NULL)
}
\arguments{
\item{files, variables, sampler_diagnostics}{Deprecated. Use
\code{\link[=read_cmdstan_csv]{read_cmdstan_csv()}} instead.}
}
\description{
Deprecated. Use \code{\link[=read_cmdstan_csv]{read_cmdstan_csv()}} instead.
}
\keyword{internal}
